package token

import (
	"errors"
	"fmt"
	commonCliUtils "github.com/jfrog/jfrog-cli-core/v2/common/cliutils"
	"github.com/jfrog/jfrog-cli-core/v2/common/commands"
	generic "github.com/jfrog/jfrog-cli-core/v2/general/token"
	coreConfig "github.com/jfrog/jfrog-cli-core/v2/utils/config"
	"github.com/jfrog/jfrog-cli-core/v2/utils/coreutils"
	"github.com/jfrog/jfrog-cli/utils/accesstoken"
	"github.com/jfrog/jfrog-cli/utils/cliutils"
	clientUtils "github.com/jfrog/jfrog-client-go/utils"
	"github.com/jfrog/jfrog-client-go/utils/errorutils"
	"github.com/jfrog/jfrog-client-go/utils/log"
	"github.com/urfave/cli"
	"os"
	"strconv"
)

func AccessTokenCreateCmd(c *cli.Context) error {
	if c.NArg() > 1 {
		return cliutils.WrongNumberOfArgumentsHandler(c)
	}

	serverDetails, err := createPlatformDetailsByFlags(c)
	if err != nil {
		return err
	}

	if err = assertAccessTokenAvailable(serverDetails); err != nil {
		return err
	}

	if err = assertScopeOptions(c); err != nil {
		return err
	}

	username := accesstoken.GetSubjectUsername(c, serverDetails)

	expiry, err := getExpiry(c)
	if err != nil {
		return err
	}

	accessTokenCreateCmd := generic.NewAccessTokenCreateCommand()
	accessTokenCreateCmd.
		SetServerDetails(serverDetails).
		SetUsername(username).
		SetProjectKey(c.String(cliutils.Project)).
		SetGroups(c.String(cliutils.Groups)).
		SetScope(c.String(cliutils.Scope)).
		SetGrantAdmin(c.Bool(cliutils.GrantAdmin)).
		SetExpiry(expiry).
		SetRefreshable(c.Bool(cliutils.Refreshable)).
		SetDescription(c.String(cliutils.Description)).
		SetAudience(c.String(cliutils.Audience)).
		SetIncludeReferenceToken(c.Bool(cliutils.Reference))
	err = commands.Exec(accessTokenCreateCmd)
	if err != nil {
		return err
	}
	resString, err := accessTokenCreateCmd.Response()
	if err != nil {
		return err
	}
	log.Output(clientUtils.IndentJson(resString))

	return nil
}

func ExchangeOidcTokenCmd(c *cli.Context) error {
	if c.NArg() < 1 {
		return cliutils.WrongNumberOfArgumentsHandler(c)
	}

	serverDetails, err := createPlatformDetailsByFlags(c)
	if err != nil {
		return err
	}

	oidcAccessTokenCreateCmd, err := CreateOidcTokenExchangeCommand(c, serverDetails)
	if err != nil {
		return err
	}

	if err = commands.ExecAndThenReportUsage(oidcAccessTokenCreateCmd); err != nil {
		return err
	}
	// Print to the console only if the specific EOT command has been executed.
	// Otherwise, it will be printed by the config command, which is unnecessary.
	oidcAccessTokenCreateCmd.PrintResponseToConsole()
	return nil
}

func CreateOidcTokenExchangeCommand(c *cli.Context, serverDetails *coreConfig.ServerDetails) (*generic.OidcTokenExchangeCommand, error) {
	oidcAccessTokenCreateCmd := generic.NewOidcTokenExchangeCommand()
	// Validate supported oidc provider type
	if err := oidcAccessTokenCreateCmd.SetProviderTypeAsString(cliutils.GetFlagOrEnvValue(c, cliutils.OidcProviderType, coreutils.OidcProviderType)); err != nil {
		return nil, err
	}

	oidcAccessTokenCreateCmd.
		SetServerDetails(serverDetails).
		// Mandatory flags
		SetProviderName(c.Args().Get(0)).
		SetOidcTokenID(getOidcTokenIdInput(c)).
		SetAudience(c.String(cliutils.OidcAudience)).
		// Optional values exported by CI servers
		SetJobId(os.Getenv(coreutils.CIJobID)).
		SetRunId(os.Getenv(coreutils.CIRunID)).
		// Values which can both be exported or explicitly set
		SetProjectKey(cliutils.GetFlagOrEnvValue(c, cliutils.Project, coreutils.Project)).
		SetApplicationKey(cliutils.GetJFrogApplicationKey(c))

	return oidcAccessTokenCreateCmd, nil
}

func createPlatformDetailsByFlags(c *cli.Context) (*coreConfig.ServerDetails, error) {
	platformDetails, err := cliutils.CreateServerDetailsWithConfigOffer(c, true, commonCliUtils.Platform)
	if err != nil {
		return nil, err
	}
	if platformDetails.Url == "" {
		return nil, errors.New("no JFrog Platform URL specified, either via the --url flag or as part of the server configuration")
	}
	return platformDetails, nil
}

func getExpiry(c *cli.Context) (*uint, error) {
	if !c.IsSet(cliutils.Expiry) {
		return nil, nil
	}

	expiryInt, err := strconv.Atoi(c.String(cliutils.Expiry))
	if err != nil {
		return nil, cliutils.PrintHelpAndReturnError(
			fmt.Sprintf("The '--%s' option must have a numeric value. ", cliutils.Expiry), c)
	}
	if expiryInt < 0 {
		return nil, cliutils.PrintHelpAndReturnError(
			fmt.Sprintf("The '--%s' option must be non-negative. ", cliutils.Expiry), c)
	}
	expiry := uint(expiryInt)
	return &expiry, nil
}

func assertScopeOptions(c *cli.Context) error {
	if c.IsSet(cliutils.Scope) && (c.IsSet(cliutils.GrantAdmin) || c.IsSet(cliutils.Groups)) {
		return cliutils.PrintHelpAndReturnError(
			fmt.Sprintf("Scope can either be provided explicitly with '--%s', or implicitly with '--%s' and '--%s'. ",
				cliutils.Scope, cliutils.GrantAdmin, cliutils.Groups), c)
	}
	return nil
}

func assertAccessTokenAvailable(serverDetails *coreConfig.ServerDetails) error {
	if serverDetails.AccessToken == "" {
		return errorutils.CheckErrorf("authenticating with access token is currently mandatory for creating access tokens")
	}
	return nil
}

// The OIDC token ID can be provided as a command line argument or as a flag or environment variable.
// Depends on the origin of the command.
// For example when used in CI/CD, the token ID is provided as a environment variable.
func getOidcTokenIdInput(c *cli.Context) string {
	oidcTokenId := c.Args().Get(1)
	if oidcTokenId == "" {
		oidcTokenId = cliutils.GetFlagOrEnvValue(c, cliutils.OidcTokenID, coreutils.OidcExchangeTokenId)
	}
	return oidcTokenId
}
